@@ -45,6 +45,7 @@
 #include <KIO/NetAccess>
 #include <KIcon>
 #include <KInputDialog>
+#include <KMenu>
 #include <KMessageBox>
 #include <KPluginFactory>
 #include <KRun>
@@ -140,6 +141,8 @@ Part::~Part()
 {
     factory()->removeClient(this);
 
+    qDeleteAll(m_previewDirList);
+
     saveSplitterSizes();
 
     m_extractFilesAction->menu()->deleteLater();
@@ -213,6 +216,8 @@ void Part::extractSelectedFilesTo(const QString& localPath)
 
 void Part::setupView()
 {
+    m_view->setContextMenuPolicy(Qt::CustomContextMenu);
+
     m_view->setModel(m_model);
 
     m_view->setSortingEnabled(true);
@@ -223,8 +228,10 @@ void Part::setupView()
             this, SLOT(selectionChanged()));
 
     //TODO: fix an actual eventhandler
-    connect(m_view, SIGNAL(itemTriggered(QModelIndex)),
-            this, SLOT(slotPreview(QModelIndex)));
+    connect(m_view, SIGNAL(activated(const QModelIndex &)),
+            this, SLOT(slotPreviewWithInternalViewer()));
+
+    connect(m_view, SIGNAL(customContextMenuRequested(QPoint)), this, SLOT(slotShowContextMenu()));
 
     connect(m_model, SIGNAL(columnsInserted(QModelIndex,int,int)),
             this, SLOT(adjustColumns()));
@@ -240,13 +247,19 @@ void Part::setupActions()
 
     m_saveAsAction = KStandardAction::saveAs(this, SLOT(slotSaveAs()), actionCollection());
 
+    m_previewChooseAppAction = actionCollection()->addAction(QLatin1String("openwith"));
+    m_previewChooseAppAction->setText(i18nc("open a file with external program", "Open &With..."));
+    m_previewChooseAppAction->setIcon(KIcon(QLatin1String("document-open")));
+    m_previewChooseAppAction->setStatusTip(i18n("Click to open the selected file with an external program"));
+    connect(m_previewChooseAppAction, SIGNAL(triggered(bool)), this, SLOT(slotPreviewWithExternalProgram()));
+
     m_previewAction = actionCollection()->addAction(QLatin1String( "preview" ));
     m_previewAction->setText(i18nc("to preview a file inside an archive", "Pre&view"));
     m_previewAction->setIcon(KIcon( QLatin1String( "document-preview-archive" )));
     m_previewAction->setStatusTip(i18n("Click to preview the selected file"));
     m_previewAction->setShortcuts(QList<QKeySequence>() << Qt::Key_Return << Qt::Key_Space);
     connect(m_previewAction, SIGNAL(triggered(bool)),
-            this, SLOT(slotPreview()));
+            this, SLOT(slotPreviewWithInternalViewer()));
 
     m_extractFilesAction = actionCollection()->addAction(QLatin1String( "extract" ));
     m_extractFilesAction->setText(i18n("E&xtract"));
@@ -292,6 +305,8 @@ void Part::updateActions()
     m_addDirAction->setEnabled(!isBusy() && isWritable);
     m_deleteFilesAction->setEnabled(!isBusy() && (m_view->selectionModel()->selectedRows().count() > 0)
                                     && isWritable);
+    m_previewChooseAppAction->setEnabled(!isBusy() && (m_view->selectionModel()->selectedRows().count() > 0)
+                                    && isWritable);
 
     QMenu *menu = m_extractFilesAction->menu();
     if (!menu) {
@@ -537,12 +552,17 @@ void Part::setFileNameFromArchive()
     emit setWindowCaption(prettyName);
 }
 
-void Part::slotPreview()
+void Part::slotPreviewWithInternalViewer()
 {
-    slotPreview(m_view->selectionModel()->currentIndex());
+    preview(m_view->selectionModel()->currentIndex(), InternalViewer);
 }
 
-void Part::slotPreview(const QModelIndex & index)
+void Part::slotPreviewWithExternalProgram()
+{
+    preview(m_view->selectionModel()->currentIndex(), ExternalProgram);
+}
+
+void Part::preview(const QModelIndex &index, PreviewMode mode)
 {
     if (!isPreviewable(index)) {
         return;
@@ -554,7 +574,10 @@ void Part::slotPreview(const QModelIndex & index)
         Kerfuffle::ExtractionOptions options;
         options[QLatin1String( "PreservePaths" )] = true;
 
-        ExtractJob *job = m_model->extractFile(entry[ InternalID ], m_previewDir.name(), options);
+        m_previewDirList.append(new KTempDir);
+        m_previewMode = mode;
+        ExtractJob *job = m_model->extractFile(entry[InternalID], m_previewDirList.last()->name(), options);
+
         registerJob(job);
         connect(job, SIGNAL(result(KJob*)),
                 this, SLOT(slotPreviewExtracted(KJob*)));
@@ -571,15 +594,28 @@ void Part::slotPreviewExtracted(KJob *job)
         const ArchiveEntry& entry =
             m_model->entryForIndex(m_view->selectionModel()->currentIndex());
 
-        QString fullName =
-            m_previewDir.name() + QLatin1Char('/') + entry[FileName].toString();
+        ExtractJob *extractJob = qobject_cast<ExtractJob*>(job);
+        Q_ASSERT(extractJob);
+        QString fullName = extractJob->destinationDirectory() + entry[FileName].toString();
 
         // Make sure a maliciously crafted archive with parent folders named ".." do
         // not cause the previewed file path to be located outside the temporary
         // directory, resulting in a directory traversal issue.
         fullName.remove(QLatin1String("../"));
 
-        ArkViewer::view(fullName, widget());
+        // TODO: get rid of m_previewMode by extending ExtractJob with a PreviewJob.
+        // This would prevent race conditions if we ever stop disabling
+        // the whole UI while extracting a file to preview it.
+        switch (m_previewMode) {
+        case InternalViewer:
+            ArkViewer::view(fullName, widget());
+            break;
+        case ExternalProgram:
+            KUrl::List list;
+            list.append(KUrl(fullName));
+            KRun::displayOpenWithDialog(list, widget(), true);
+            break;
+        }
     } else {
         KMessageBox::error(widget(), job->errorString());
     }
@@ -721,7 +757,6 @@ void Part::slotExtractionDone(KJob* job)
         }
 
         if (ArkSettings::openDestinationFolderAfterExtraction()) {
-
             KUrl destinationDirectory(extractJob->destinationDirectory());
             destinationDirectory.cleanPath();
 
@@ -913,4 +948,10 @@ void Part::slotSaveAs()
     }
 }
 
+void Part::slotShowContextMenu()
+{
+    KMenu *popup = static_cast<KMenu *>(factory()->container(QLatin1String("context_menu"), this));
+    popup->popup(QCursor::pos());
+}
+
 } // namespace Ark
